/**
 * <code>FlxEmitter</code> is a lightweight particle emitter.<br>
 * It can be used for one-time explosions or for<br>
 * continuous fx like rain and fire.  <code>FlxEmitter</code><br>
 * is not optimized or anything; all it does is launch<br>
 * <code>FlxParticle</code> objects out at set intervals<br>
 * by setting their positions and velocities accordingly.<br>
 * It is easy to use and relatively efficient,<br>
 * relying on <code>FlxGroup</code>'s RECYCLE POWERS.<br>
 * <br>
 * v1.0 Initial version
 * 
 * @version 1.0 - 17/09/2014
 * @author ratalaika / ratalaikaGames
 * @author	Adam Atomic
 * @class Flixel.FlxEmitter
 */

/**
 * Creates a new <code>FlxEmitter</code> object at a specific position. Does NOT automatically generate or attach particles!
 * 
 * @param X
 *            The X position of the emitter.
 * @param Y
 *            The Y position of the emitter.
 * @param Size
 *            Optional, specifies a maximum capacity for this emitter.
 * @constructor
 */
Flixel.FlxEmitter = function(X, Y, Size)
{
	Flixel.FlxEmitter.parent.constructor.apply(this, [ Size ]);

	this.x = X;
	this.y = Y;
	this.width = 0;
	this.height = 0;
	this.minParticleSpeed = new Flixel.FlxPoint(-100, -100);
	this.maxParticleSpeed = new Flixel.FlxPoint(100, 100);
	this.minRotation = -360;
	this.maxRotation = 360;
	this.gravity = 0;
	this._particleClass = new Flixel.FlxParticle();
	this.particleDrag = new Flixel.FlxPoint();
	this.frequency = 0.1;
	this.lifespan = 3;
	this.bounce = 0;
	this._quantity = 0;
	this._counter = 0;
	this._explode = true;
	this.on = false;
	this._point = new Flixel.FlxPoint();
	this._waitForKill = false;

};
extend(Flixel.FlxEmitter, Flixel.FlxGroup);

/**
 * The X position of the top left corner of the emitter in world space.
 */
Flixel.FlxEmitter.prototype.x = 0;
/**
 * The Y position of the top left corner of emitter in world space.
 */
Flixel.FlxEmitter.prototype.y = 0;
/**
 * The width of the emitter. Particles can be randomly generated from anywhere within this box.
 */
Flixel.FlxEmitter.prototype.width = 0;
/**
 * The height of the emitter. Particles can be randomly generated from anywhere within this box.
 */
Flixel.FlxEmitter.prototype.height = 0;
/**
 * The minimum possible velocity of a particle. The default value is (-100,-100).
 */
Flixel.FlxEmitter.prototype.minParticleSpeed = null;
/**
 * The maximum possible velocity of a particle. The default value is (100,100).
 */
Flixel.FlxEmitter.prototype.maxParticleSpeed = null;
/**
 * The X and Y drag component of particles launched from the emitter.
 */
Flixel.FlxEmitter.prototype.particleDrag = null;
/**
 * The minimum possible angular velocity of a particle. The default value is -360. NOTE: rotating particles are more expensive to draw than non-rotating ones!
 */
Flixel.FlxEmitter.prototype.minRotation = 0;
/**
 * The maximum possible angular velocity of a particle. The default value is 360. NOTE: rotating particles are more expensive to draw than non-rotating ones!
 */
Flixel.FlxEmitter.prototype.maxRotation = 0;
/**
 * Sets the <code>acceleration.y</code> member of each particle to this value on launch.
 */
Flixel.FlxEmitter.prototype.gravity = 0;
/**
 * Determines whether the emitter is currently emitting particles. It is totally safe to directly toggle this.
 */
Flixel.FlxEmitter.prototype.on = false;
/**
 * How often a particle is emitted (if emitter is started with Explode == false).
 */
Flixel.FlxEmitter.prototype.frequency = 0;
/**
 * How long each particle lives once it is emitted. Set lifespan to 'zero' for particles to live forever.
 */
Flixel.FlxEmitter.prototype.lifespan = 0;
/**
 * How much each particle should bounce. 1 = full bounce, 0 = no bounce.
 */
Flixel.FlxEmitter.prototype.bounce = 0;
/**
 * Internal variable for tracking the class to create when generating particles.
 */
Flixel.FlxEmitter.prototype._particleClass = null;
/**
 * Internal helper for deciding how many particles to launch.
 */
Flixel.FlxEmitter.prototype._quantity = 0;
/**
 * Internal helper for the style of particle emission (all at once, or one at a time).
 */
Flixel.FlxEmitter.prototype._explode = false;
/**
 * Internal helper for deciding when to launch particles or kill them.
 */
Flixel.FlxEmitter.prototype._timer = 0;
/**
 * Internal counter for figuring out how many particles to launch.
 */
Flixel.FlxEmitter.prototype._counter = 0;
/**
 * Internal point object, handy for reusing for memory mgmt purposes.
 */
Flixel.FlxEmitter.prototype._point = null;
/**
 * Internal helper for automatic call the kill() method
 */
Flixel.FlxEmitter.prototype._waitForKill = false;

/**
 * Clean up memory.
 */
Flixel.FlxEmitter.prototype.destroy = function()
{
	this.minParticleSpeed = null;
	this.maxParticleSpeed = null;
	this.particleDrag = null;
	this._particleClass = null;
	this._point = null;

	Flixel.FlxEmitter.parent.destroy.apply(this);
};

/**
 * This function generates a new array of particle sprites to attach to the emitter.
 * 
 * @param Graphics
 *            If you opted to not pre-configure an array of FlxParticle objects, you can simply pass in a particle image or sprite sheet.
 * @param Quantity
 *            The number of particles to generate when using the "create from image" option.
 * @param BakedRotations
 *            How many frames of baked rotation to use (boosts performance). Set to zero to not use baked rotations.
 * @param Multiple
 *            Whether the image in the Graphics param is a single particle or a bunch of particles (if it's a bunch, they need to be square!).
 * @param Collide
 *            Whether the particles should be flagged as not 'dead' (non-colliding particles are higher performance). 0 means no collisions, 0-1 controls scale of particle's bounding box.
 * 
 * @return This FlxEmitter instance (nice for chaining stuff together, if you're into that).
 */
Flixel.FlxEmitter.prototype.makeParticles = function(Graphics, Quantity, BakedRotations, Multiple, Collide)
{
	BakedRotations = BakedRotations || 16;
	Collide = Collide || 0.8;

	this._maxSize = Quantity || 50;

	var totalFrames = 1;
	if (Multiple) {
		var sprite = new Flixel.FlxSprite();
		sprite.loadGraphic(Graphics, true);
		totalFrames = sprite.frames;
		sprite.destroy();
	}

	var randomFrame = 0;
	var particle = null;
	var i = 0;
	while (i < Quantity) {
		particle = new this._particleClass.constructor();

		if (Multiple) {
			randomFrame = Flixel.FlxG.random() * totalFrames;
			if (BakedRotations > 0)
				particle.loadRotatedGraphic(Graphics, BakedRotations, randomFrame);
			else {
				particle.loadGraphic(Graphics, true);
				particle.setFrame(randomFrame);
			}
		} else {
			if (BakedRotations > 0)
				particle.loadRotatedGraphic(Graphics, BakedRotations);
			else
				particle.loadGraphic(Graphics);
		}
		if (Collide > 0) {
			particle.width *= Collide;
			particle.height *= Collide;
			particle.centerOffsets();
		} else
			particle.allowCollisions = Flixel.FlxObject.NONE;
		particle.exists = false;
		this.add(particle);
		i++;
	}
	return this;
};

/**
 * Called automatically by the game loop, decides when to launch particles and when to "die".
 */
Flixel.FlxEmitter.prototype.update = function()
{
	if (this.on) {
		if (this._explode) {
			this.on = false;
			this._waitForKill = true;
			var i = 0;
			var l = this._quantity;
			if ((l <= 0) || (l > this.length))
				l = this.length;
			while (i < l) {
				this.emitParticle();
				i++;
			}
			this._quantity = 0;
		} else {
			// Spawn a particle per frame
			if (this.frequency <= 0) {
				this.emitParticle();
				if ((this._quantity > 0) && (++this._counter >= this._quantity)) {
					this.on = false;
					this._waitForKill = true;
					this._quantity = 0;
				}
			} else {
				this._timer += Flixel.FlxG.elapsed;
				while ((this.frequency > 0) && (this._timer > this.frequency) && this.on) {
					this._timer -= this.frequency;
					this.emitParticle();
					if ((this._quantity > 0) && (++this._counter >= this._quantity)) {
						this.on = false;
						this._waitForKill = true;
						this._quantity = 0;
					}
				}
			}
		}
	}

	Flixel.FlxEmitter.parent.update.apply(this);
};

/**
 * Call this function to turn off all the particles and the emitter.
 */
Flixel.FlxEmitter.prototype.kill = function()
{
	this.on = false;
	this._waitForKill = false;
	Flixel.FlxEmitter.parent.kill.apply(this);
};

/**
 * Call this function to start emitting particles.
 * 
 * @param Explode
 *            Whether the particles should all burst out at once.
 * @param Lifespan
 *            How long each particle lives once emitted. 0 = forever.
 * @param Frequency
 *            Ignored if Explode is set to true. Frequency is how often to emit a particle. 0 = never emit, 0.1 = 1 particle every 0.1 seconds, 5 = 1 particle every 5 seconds.
 * @param Quantity
 *            How many particles to launch. 0 = "all of the particles".
 */
Flixel.FlxEmitter.prototype.start = function(Explode, Lifespan, Frequency, Quantity)
{
	Explode = (Explode === undefined) ? true : Explode;
	Lifespan = Lifespan || 0;
	Frequency = Frequency || 0.1;
	Quantity = Quantity || 0;

	this.revive();
	this.visible = true;
	this.on = true;

	this._explode = Explode;
	this.lifespan = Lifespan;
	this.frequency = Frequency;
	this._quantity += Quantity;

	this._counter = 0;
	this._timer = 0;

	this._waitForKill = false;
};

/**
 * This function can be used both internally and externally to emit the next particle.
 */
Flixel.FlxEmitter.prototype.emitParticle = function()
{
	var particle = this.recycle(this._particleClass);
	particle.lifespan = this.lifespan;
	particle.elasticity = this.bounce;
	particle.reset(this.x - (particle.width >> 1) + Flixel.FlxG.random() * this.width, this.y - (particle.height >> 1) + Flixel.FlxG.random() * this.height);
	particle.visible = true;

	if (this.minParticleSpeed.x != this.maxParticleSpeed.x)
		particle.velocity.x = this.minParticleSpeed.x + Flixel.FlxG.random() * (this.maxParticleSpeed.x - this.minParticleSpeed.x);
	else
		particle.velocity.x = this.minParticleSpeed.x;
	if (this.minParticleSpeed.y != this.maxParticleSpeed.y)
		particle.velocity.y = this.minParticleSpeed.y + Flixel.FlxG.random() * (this.maxParticleSpeed.y - this.minParticleSpeed.y);
	else
		particle.velocity.y = this.minParticleSpeed.y;
	particle.acceleration.y = this.gravity;

	if (this.minRotation != this.maxRotation)
		particle.angularVelocity = this.minRotation + Flixel.FlxG.random() * (this.maxRotation - this.minRotation);
	else
		particle.angularVelocity = this.minRotation;
	if (particle.angularVelocity !== 0)
		particle.angle = Flixel.FlxG.random() * 360 - 180;

	particle.drag.x = this.particleDrag.x;
	particle.drag.y = this.particleDrag.y;
	particle.onEmit();
};

/**
 * Set your own particle class type here. The custom class must extend <code>FlxParticle</code>. Default is <code>FlxParticle</code>.
 */
Flixel.FlxEmitter.prototype.getParticleClass = function()
{
	return this._particleClass;
};

Flixel.FlxEmitter.prototype.setParticleClass = function(value)
{
	var testParticle = new value.constructor();
	if (testParticle instanceof Flixel.FlxParticle) {
		this._particleClass = value;
	} else {
		Flixel.FlxG.log("ERROR: " + Flixel.FlxU.getClassName(testParticle, true) + " must extend FlxParticle in order to be used in a FlxEmitter.");
	}
};

/**
 * A more compact way of setting the width and height of the emitter.
 * 
 * @param Width
 *            The desired width of the emitter (particles are spawned randomly within these dimensions).
 * @param Height
 *            The desired height of the emitter.
 */
Flixel.FlxEmitter.prototype.setSize = function(Width, Height)
{
	this.width = Width;
	this.height = Height;
};

/**
 * A more compact way of setting the X velocity range of the emitter.
 * 
 * @param Min
 *            The minimum value for this range.
 * @param Max
 *            The maximum value for this range.
 */
Flixel.FlxEmitter.prototype.setXSpeed = function(Min, Max)
{
	this.minParticleSpeed.x = Min || 0;
	this.maxParticleSpeed.x = Max || 0;
};

/**
 * A more compact way of setting the Y velocity range of the emitter.
 * 
 * @param Min
 *            The minimum value for this range.
 * @param Max
 *            The maximum value for this range.
 */
Flixel.FlxEmitter.prototype.setYSpeed = function(Min, Max)
{
	this.minParticleSpeed.y = Min || 0;
	this.maxParticleSpeed.y = Max || 0;
};

/**
 * A more compact way of setting the angular velocity constraints of the emitter.
 * 
 * @param Min
 *            The minimum value for this range.
 * @param Max
 *            The maximum value for this range.
 */
Flixel.FlxEmitter.prototype.setRotation = function(Min, Max)
{
	this.minRotation = Min || 0;
	this.maxRotation = Max || 0;
};

/**
 * Change the emitter's midpoint to match the midpoint of a <code>FlxObject</code>.
 * 
 * @param Object
 *            The <code>FlxObject</code> that you want to sync up with.
 */
Flixel.FlxEmitter.prototype.at = function(Object)
{
	Object.getMidpoint(this._point);
	this.x = this._point.x - (this.width >> 1);
	this.y = this._point.y - (this.height >> 1);
};

/**
 * Returns the class name.
 */
Flixel.FlxEmitter.prototype.toString = function()
{
	return "FlxEmitter";
};